@@ -2,17 +2,19 @@
 #   Symbol Table
 #
 
+from __future__ import absolute_import
+
 import copy
 import re
-from Errors import warning, error, InternalError
-from StringEncoding import EncodedString
-import Options, Naming
-import PyrexTypes
-from PyrexTypes import py_object_type, unspecified_type
-from TypeSlots import \
+from .Errors import warning, error, InternalError
+from .StringEncoding import EncodedString
+from . import Options, Naming
+from . import PyrexTypes
+from .PyrexTypes import py_object_type, unspecified_type
+from .TypeSlots import \
     pyfunction_signature, pymethod_signature, \
     get_special_method_signature, get_property_accessor_signature
-import Code
+from . import Code
 import __builtin__ as builtins
 
 iso_c99_keywords = set(
@@ -23,6 +25,7 @@ iso_c99_keywords = set(
     'volatile', 'while',
     '_Bool', '_Complex'', _Imaginary', 'inline', 'restrict'])
 
+
 def c_safe_identifier(cname):
     # There are some C limitations on struct entry names.
     if ((cname[:2] == '__'
@@ -351,7 +354,7 @@ class Scope(object):
                      'cfunc_entries',
                      'c_class_entries'):
             self_entries = getattr(self, attr)
-            names = set([e.name for e in self_entries])
+            names = set(e.name for e in self_entries)
             for entry in getattr(other, attr):
                 if (entry.used or merge_unused) and entry.name not in names:
                     self_entries.append(entry)
@@ -404,7 +407,7 @@ class Scope(object):
         """ Return the module-level scope containing this scope. """
         return self.outer_scope.builtin_scope()
 
-    def declare(self, name, cname, type, pos, visibility, shadow = 0, is_type = 0):
+    def declare(self, name, cname, type, pos, visibility, shadow = 0, is_type = 0, create_wrapper = 0):
         # Create new entry, and add to dictionary if
         # name is not None. Reports a warning if already
         # declared.
@@ -421,6 +424,7 @@ class Scope(object):
                 error(pos, "'%s' redeclared " % name)
         entry = Entry(name, cname, type, pos = pos)
         entry.in_cinclude = self.in_cinclude
+        entry.create_wrapper = create_wrapper
         if name:
             entry.qualified_name = self.qualify_name(name)
 #            if name in entries and self.is_cpp():
@@ -431,7 +435,7 @@ class Scope(object):
                 entries[name] = entry
 
         if type.is_memoryviewslice:
-            import MemoryView
+            from . import MemoryView
             entry.init = MemoryView.memslice_entry_init
 
         entry.scope = self
@@ -441,14 +445,14 @@ class Scope(object):
     def qualify_name(self, name):
         return EncodedString("%s.%s" % (self.qualified_name, name))
 
-    def declare_const(self, name, type, value, pos, cname = None, visibility = 'private', api = 0):
+    def declare_const(self, name, type, value, pos, cname = None, visibility = 'private', api = 0, create_wrapper = 0):
         # Add an entry for a named constant.
         if not cname:
             if self.in_cinclude or (visibility == 'public' or api):
                 cname = name
             else:
                 cname = self.mangle(Naming.enum_prefix, name)
-        entry = self.declare(name, cname, type, pos, visibility)
+        entry = self.declare(name, cname, type, pos, visibility, create_wrapper = create_wrapper)
         entry.is_const = 1
         entry.value_node = value
         return entry
@@ -569,7 +573,7 @@ class Scope(object):
                     entry.type.scope.declare_inherited_cpp_attributes(base_class.scope)
         if entry.type.scope:
             declare_inherited_attributes(entry, base_classes)
-            entry.type.scope.declare_var(name="this", cname="this", type=PyrexTypes.CPtrType(entry.type), pos=entry.pos)
+            scope.declare_var(name="this", cname="this", type=PyrexTypes.CPtrType(entry.type), pos=entry.pos)
         if self.is_cpp_class_scope:
             entry.type.namespace = self.outer_scope.lookup(self.name).type
         return entry
@@ -585,7 +589,7 @@ class Scope(object):
                 entry.name, entry.visibility))
 
     def declare_enum(self, name, pos, cname, typedef_flag,
-            visibility = 'private', api = 0):
+            visibility = 'private', api = 0, create_wrapper = 0):
         if name:
             if not cname:
                 if self.in_cinclude or (visibility == 'public' or api):
@@ -597,6 +601,7 @@ class Scope(object):
             type = PyrexTypes.c_anon_enum_type
         entry = self.declare_type(name, type, pos, cname = cname,
             visibility = visibility, api = api)
+        entry.create_wrapper = create_wrapper
         entry.enum_values = []
         self.sue_entries.append(entry)
         return entry
@@ -801,7 +806,7 @@ class Scope(object):
         return PyrexTypes.best_match(operands, function.all_alternatives())
 
     def lookup_operator_for_types(self, pos, operator, types):
-        from Nodes import Node
+        from .Nodes import Node
         class FakeOperand(Node):
             pass
         operands = [FakeOperand(pos, type=type) for type in types]
@@ -823,7 +828,7 @@ class Scope(object):
         return 0
 
     def infer_types(self):
-        from TypeInference import get_type_inferer
+        from .TypeInference import get_type_inferer
         get_type_inferer().infer_types(self)
 
     def is_cpp(self):
@@ -900,6 +905,7 @@ class BuiltinScope(Scope):
             var_entry.is_variable = 1
             var_entry.is_builtin = 1
             var_entry.utility_code = utility_code
+            var_entry.scope = entry.scope
             entry.as_variable = var_entry
         return entry
 
@@ -983,7 +989,7 @@ class ModuleScope(Scope):
     # included_files       [string]           Cython sources included with 'include'
     # pxd_file_loaded      boolean            Corresponding .pxd file has been processed
     # cimported_modules    [ModuleScope]      Modules imported with cimport
-    # types_imported       {PyrexType : 1}    Set of types for which import code generated
+    # types_imported       {PyrexType}        Set of types for which import code generated
     # has_import_star      boolean            Module contains import *
     # cpp                  boolean            Compiling a C++ file
     # is_cython_builtin    boolean            Is this the Cython builtin scope (or a child scope)
@@ -994,7 +1000,7 @@ class ModuleScope(Scope):
     is_cython_builtin = 0
 
     def __init__(self, name, parent_module, context):
-        import Builtin
+        from . import Builtin
         self.parent_module = parent_module
         outer_scope = Builtin.builtin_scope
         Scope.__init__(self, name, outer_scope, parent_module)
@@ -1020,7 +1026,7 @@ class ModuleScope(Scope):
         self.type_names = dict(outer_scope.type_names)
         self.pxd_file_loaded = 0
         self.cimported_modules = []
-        self.types_imported = {}
+        self.types_imported = set()
         self.included_files = []
         self.has_extern_class = 0
         self.cached_builtins = []
@@ -1080,13 +1086,19 @@ class ModuleScope(Scope):
             entry.name = name
         return entry
 
-    def find_module(self, module_name, pos):
+    def find_module(self, module_name, pos, relative_level=-1):
         # Find a module in the import namespace, interpreting
         # relative imports relative to this module's parent.
         # Finds and parses the module's .pxd file if the module
         # has not been referenced before.
-        return self.global_scope().context.find_module(
-            module_name, relative_to = self.parent_module, pos = pos)
+        module_scope = self.global_scope()
+        if relative_level is not None and relative_level > 0:
+            # merge current absolute module name and relative import name into qualified name
+            current_module = module_scope.qualified_name.split('.')
+            base_package = current_module[:-relative_level]
+            module_name = '.'.join(base_package + module_name.split('.'))
+        return module_scope.context.find_module(
+            module_name, relative_to=None if relative_level == 0 else self.parent_module, pos=pos)
 
     def find_submodule(self, name):
         # Find and return scope for a submodule of this module,
@@ -1454,7 +1466,7 @@ class ModuleScope(Scope):
         # variable entry attached to it. For the variable entry,
         # we use a read-only C global variable whose name is an
         # expression that refers to the type object.
-        import Builtin
+        from . import Builtin
         var_entry = Entry(name = entry.name,
             type = Builtin.type_type,
             pos = entry.pos,
@@ -1468,7 +1480,7 @@ class ModuleScope(Scope):
         return self.cpp
 
     def infer_types(self):
-        from TypeInference import PyObjectTypeInferer
+        from .TypeInference import PyObjectTypeInferer
         PyObjectTypeInferer().infer_types(self)
 
 
@@ -1768,6 +1780,7 @@ class CClassScope(ClassScope):
     #  method_table_cname    string
     #  getset_table_cname    string
     #  has_pyobject_attrs    boolean  Any PyObject attributes?
+    #  has_memoryview_attrs  boolean  Any memory view attributes?
     #  has_cyclic_pyobject_attrs    boolean  Any PyObject attributes that may need GC?
     #  property_entries      [Entry]
     #  defined               boolean  Defined in .pxd file
@@ -1777,6 +1790,7 @@ class CClassScope(ClassScope):
     is_c_class_scope = 1
 
     has_pyobject_attrs = False
+    has_memoryview_attrs = False
     has_cyclic_pyobject_attrs = False
     defined = False
     implemented = False
@@ -1850,7 +1864,9 @@ class CClassScope(ClassScope):
             entry = self.declare(name, cname, type, pos, visibility)
             entry.is_variable = 1
             self.var_entries.append(entry)
-            if type.is_pyobject and name != '__weakref__':
+            if type.is_memoryviewslice:
+                self.has_memoryview_attrs = True
+            elif type.is_pyobject and name != '__weakref__':
                 self.has_pyobject_attrs = True
                 if (not type.is_builtin_type
                         or not type.scope or type.scope.needs_gc()):
@@ -1931,11 +1947,12 @@ class CClassScope(ClassScope):
         if get_special_method_signature(name) and not self.parent_type.is_builtin_type:
             error(pos, "Special methods must be declared with 'def', not 'cdef'")
         args = type.args
-        if not args:
-            error(pos, "C method has no self argument")
-        elif not self.parent_type.assignable_from(args[0].type):
-            error(pos, "Self argument (%s) of C method '%s' does not match parent type (%s)" %
-                  (args[0].type, name, self.parent_type))
+        if not type.is_static_method:
+            if not args:
+                error(pos, "C method has no self argument")
+            elif not self.parent_type.assignable_from(args[0].type):
+                error(pos, "Self argument (%s) of C method '%s' does not match parent type (%s)" %
+                      (args[0].type, name, self.parent_type))
         entry = self.lookup_here(name)
         if cname is None:
             cname = c_safe_identifier(name)
@@ -2090,7 +2107,8 @@ class CppClassScope(Scope):
             entry = self.declare(name, cname, type, pos, visibility)
         entry.is_variable = 1
         if type.is_cfunction and self.type:
-            entry.func_cname = "%s::%s" % (self.type.declaration_code(""), cname)
+            if not self.type.templates or not any(T.is_fused for T in self.type.templates):
+                entry.func_cname = "%s::%s" % (self.type.declaration_code(""), cname)
         if name != "this" and (defining or name != "<init>"):
             self.var_entries.append(entry)
         if type.is_pyobject and not allow_pyobject:
@@ -2098,40 +2116,10 @@ class CppClassScope(Scope):
                 "C++ class member cannot be a Python object")
         return entry
 
-    def check_base_default_constructor(self, pos):
-        # Look for default constructors in all base classes.
-        if self.default_constructor is None:
-            entry = self.lookup(self.name)
-            if not entry.type.base_classes:
-                self.default_constructor = True
-                return
-            for base_class in entry.type.base_classes:
-                if base_class is PyrexTypes.error_type:
-                    continue
-                temp_entry = base_class.scope.lookup_here("<init>")
-                found = False
-                if temp_entry is None:
-                    continue
-                for alternative in temp_entry.all_alternatives():
-                    type = alternative.type
-                    if type.is_ptr:
-                        type = type.base_type
-                    if not type.args:
-                        found = True
-                        break
-                if not found:
-                    self.default_constructor = temp_entry.scope.name
-                    error(pos, "no matching function for call to " \
-                            "%s::%s()" % (temp_entry.scope.name, temp_entry.scope.name))
-        elif not self.default_constructor:
-            error(pos, "no matching function for call to %s::%s()" %
-                  (self.default_constructor, self.default_constructor))
-
     def declare_cfunction(self, name, type, pos,
                           cname = None, visibility = 'extern', api = 0, in_pxd = 0,
                           defining = 0, modifiers = (), utility_code = None):
         if name in (self.name.split('::')[-1], '__init__') and cname is None:
-            self.check_base_default_constructor(pos)
             cname = self.type.cname
             name = '<init>'
             type.return_type = PyrexTypes.InvisibleVoidType()
@@ -2173,8 +2161,9 @@ class CppClassScope(Scope):
                                            utility_code = base_entry.utility_code)
             entry.is_inherited = 1
 
-    def specialize(self, values):
+    def specialize(self, values, type_entry):
         scope = CppClassScope(self.name, self.outer_scope)
+        scope.type = type_entry
         for entry in self.entries.values():
             if entry.is_type:
                 scope.declare_type(entry.name,
