@@ -2,12 +2,10 @@
 GDB extension that adds Cython support.
 """
 
-from __future__ import with_statement
+from __future__ import print_function
 
-import os
 import sys
 import textwrap
-import operator
 import traceback
 import functools
 import itertools
@@ -16,8 +14,8 @@ import collections
 import gdb
 
 try:
-  from lxml import etree
-  have_lxml = True
+    from lxml import etree
+    have_lxml = True
 except ImportError:
     have_lxml = False
     try:
@@ -56,6 +54,7 @@ PythonObject = 'PythonObject'
 _data_types = dict(CObject=CObject, PythonObject=PythonObject)
 _filesystemencoding = sys.getfilesystemencoding() or 'UTF-8'
 
+
 # decorators
 
 def dont_suppress_errors(function):
@@ -70,6 +69,7 @@ def dont_suppress_errors(function):
 
     return wrapper
 
+
 def default_selected_gdb_frame(err=True):
     def decorator(function):
         @functools.wraps(function)
@@ -86,6 +86,7 @@ def default_selected_gdb_frame(err=True):
         return wrapper
     return decorator
 
+
 def require_cython_frame(function):
     @functools.wraps(function)
     @require_running_program
@@ -97,6 +98,7 @@ def require_cython_frame(function):
         return function(self, *args, **kwargs)
     return wrapper
 
+
 def dispatch_on_frame(c_command, python_command=None):
     def decorator(function):
         @functools.wraps(function)
@@ -117,6 +119,7 @@ def dispatch_on_frame(c_command, python_command=None):
         return wrapper
     return decorator
 
+
 def require_running_program(function):
     @functools.wraps(function)
     def wrapper(*args, **kwargs):
@@ -154,8 +157,6 @@ class CythonModule(object):
         self.lineno_c2cy = {}
         self.functions = {}
 
-    def qualified_name(self, varname):
-        return '.'.join(self.name, varname)
 
 class CythonVariable(object):
 
@@ -166,6 +167,7 @@ class CythonVariable(object):
         self.type = type
         self.lineno = int(lineno)
 
+
 class CythonFunction(CythonVariable):
     def __init__(self,
                  module,
@@ -174,7 +176,8 @@ class CythonFunction(CythonVariable):
                  pf_cname,
                  qualified_name,
                  lineno,
-                 type=CObject):
+                 type=CObject,
+                 is_initmodule_function="False"):
         super(CythonFunction, self).__init__(name,
                                              cname,
                                              qualified_name,
@@ -182,6 +185,7 @@ class CythonFunction(CythonVariable):
                                              lineno)
         self.module = module
         self.pf_cname = pf_cname
+        self.is_initmodule_function = is_initmodule_function == "True"
         self.locals = {}
         self.arguments = []
         self.step_into_functions = set()
@@ -243,7 +247,8 @@ class CythonBase(object):
             pyframeobject = libpython.Frame(frame).get_pyop()
 
             if not pyframeobject:
-                raise gdb.GdbError('Unable to read information on python frame')
+                raise gdb.GdbError(
+                            'Unable to read information on python frame')
 
             filename = pyframeobject.filename()
             lineno = pyframeobject.current_line_num()
@@ -256,7 +261,7 @@ class CythonBase(object):
                 filename = None
                 lineno = 0
             else:
-                filename = symbol_and_line_obj.symtab.filename
+                filename = symbol_and_line_obj.symtab.fullname()
                 lineno = symbol_and_line_obj.line
                 if pygments:
                     lexer = pygments.lexers.CLexer(stripall=False)
@@ -279,7 +284,7 @@ class CythonBase(object):
         if self.is_cython_function(frame) or self.is_python_function(frame):
             return True
         elif older_frame and self.is_cython_function(older_frame):
-            # direct C function call from a Cython function
+            # check for direct C function call from a Cython function
             cython_func = self.get_cython_function(older_frame)
             return name in cython_func.step_into_functions
 
@@ -299,7 +304,7 @@ class CythonBase(object):
         try:
             source_desc, lineno = self.get_source_desc(frame)
         except NoFunctionNameInFrameError:
-            print '#%-2d Unknown Frame (compile with -g)' % index
+            print('#%-2d Unknown Frame (compile with -g)' % index)
             return
 
         if not is_c and self.is_python_function(frame):
@@ -333,15 +338,15 @@ class CythonBase(object):
             func_address = int(str(gdb_value.address).split()[0], 0)
 
         a = ', '.join('%s=%s' % (name, val) for name, val in func_args)
-        print '#%-2d 0x%016x in %s(%s)' % (index, func_address, func_name, a),
+        sys.stdout.write('#%-2d 0x%016x in %s(%s)' % (index, func_address, func_name, a))
 
         if source_desc.filename is not None:
-            print 'at %s:%s' % (source_desc.filename, lineno),
+            sys.stdout.write(' at %s:%s' % (source_desc.filename, lineno))
 
-        print
+        sys.stdout.write('\n')
 
         try:
-            print '    ' + source_desc.get_source(lineno)
+            sys.stdout.write('    ' + source_desc.get_source(lineno))
         except gdb.GdbError:
             pass
 
@@ -383,16 +388,24 @@ class CythonBase(object):
             typename = '(%s) ' % (value.type,)
 
         if max_name_length is None:
-            print '%s%s = %s%s' % (prefix, name, typename, value)
+            print('%s%s = %s%s' % (prefix, name, typename, value))
         else:
-            print '%s%-*s = %s%s' % (prefix, max_name_length, name, typename,
-                                     value)
+            print('%s%-*s = %s%s' % (prefix, max_name_length, name, typename, value))
 
     def is_initialized(self, cython_func, local_name):
+        cyvar = cython_func.locals[local_name]
         cur_lineno = self.get_cython_lineno()
-        return (local_name in cython_func.arguments or
-                (local_name in cython_func.locals and
-                 cur_lineno > cython_func.locals[local_name].lineno))
+
+        if '->' in cyvar.cname:
+            # Closed over free variable
+            if cur_lineno > cython_func.lineno:
+                if cyvar.type == PythonObject:
+                    return long(gdb.parse_and_eval(cyvar.cname))
+                return True
+            return False
+
+        return cur_lineno > cyvar.lineno
+
 
 class SourceFileDescriptor(object):
     def __init__(self, filename, lexer, formatter=None):
@@ -466,6 +479,7 @@ class CyGDBError(gdb.GdbError):
         args = args or (self.msg,)
         super(CyGDBError, self).__init__(*args)
 
+
 class NoCythonFunctionInFrameError(CyGDBError):
     """
     raised when the user requests the current cython function, which is
@@ -473,6 +487,7 @@ class NoCythonFunctionInFrameError(CyGDBError):
     """
     msg = "Current function is a function cygdb doesn't know about"
 
+
 class NoFunctionNameInFrameError(NoCythonFunctionInFrameError):
     """
     raised when the name of the C function could not be determined
@@ -496,26 +511,31 @@ class CythonParameter(gdb.Parameter):
         if default is not None:
             self.value = default
 
-    def __nonzero__(self):
+    def __bool__(self):
         return bool(self.value)
 
-    __bool__ = __nonzero__ # python 3
+    __nonzero__ = __bool__  # Python 2
+
+
 
 class CompleteUnqualifiedFunctionNames(CythonParameter):
     """
     Have 'cy break' complete unqualified function or method names.
     """
 
+
 class ColorizeSourceCode(CythonParameter):
     """
     Tell cygdb whether to colorize source code.
     """
 
+
 class TerminalBackground(CythonParameter):
     """
     Tell cygdb about the user's terminal background (light or dark).
     """
 
+
 class CythonParameters(object):
     """
     Simple container class that might get more functionality in the distant
@@ -585,6 +605,7 @@ class CyCy(CythonCommand):
         cy bt / cy backtrace
         cy list
         cy print
+        cy set
         cy locals
         cy globals
         cy exec
@@ -600,6 +621,7 @@ class CyCy(CythonCommand):
                                             completer_class, prefix=True)
 
         commands = dict(
+            # GDB commands
             import_ = CyImport.register(),
             break_ = CyBreak.register(),
             step = CyStep.register(),
@@ -617,12 +639,16 @@ class CyCy(CythonCommand):
             globals = CyGlobals.register(),
             exec_ = libpython.FixGdbCommand('cy exec', '-cy-exec'),
             _exec = CyExec.register(),
+            set = CySet.register(),
+
+            # GDB functions
             cy_cname = CyCName('cy_cname'),
             cy_cvalue = CyCValue('cy_cvalue'),
             cy_lineno = CyLine('cy_lineno'),
+            cy_eval = CyEval('cy_eval'),
         )
 
-        for command_name, command in commands.iteritems():
+        for command_name, command in commands.items():
             command.cy = self
             setattr(self, command_name, command)
 
@@ -658,9 +684,8 @@ class CyImport(CythonCommand):
         for arg in string_to_argv(args):
             try:
                 f = open(arg)
-            except OSError, e:
-                raise gdb.GdbError('Unable to open file %r: %s' %
-                                                (args, e.args[1]))
+            except OSError as e:
+                raise gdb.GdbError('Unable to open file %r: %s' % (args, e.args[1]))
 
             t = etree.parse(f)
 
@@ -706,8 +731,6 @@ class CyImport(CythonCommand):
                     for c_lineno in c_linenos:
                         cython_module.lineno_c2cy[c_lineno] = cython_lineno
 
-        self.cy.step.init_breakpoints()
-
 
 class CyBreak(CythonCommand):
     """
@@ -749,24 +772,30 @@ class CyBreak(CythonCommand):
             breakpoint = '%s:%s' % (cython_module.c_filename, c_lineno)
             gdb.execute('break ' + breakpoint)
         else:
-            raise GdbError("Not a valid line number. "
-                           "Does it contain actual code?")
+            raise gdb.GdbError("Not a valid line number. "
+                               "Does it contain actual code?")
 
     def _break_funcname(self, funcname):
         func = self.cy.functions_by_qualified_name.get(funcname)
+
+        if func and func.is_initmodule_function:
+            func = None
+
         break_funcs = [func]
 
         if not func:
-            funcs = self.cy.functions_by_name.get(funcname)
+            funcs = self.cy.functions_by_name.get(funcname) or []
+            funcs = [f for f in funcs if not f.is_initmodule_function]
+
             if not funcs:
                 gdb.execute('break ' + funcname)
                 return
 
             if len(funcs) > 1:
                 # multiple functions, let the user pick one
-                print 'There are multiple such functions:'
+                print('There are multiple such functions:')
                 for idx, func in enumerate(funcs):
-                    print '%3d) %s' % (idx, func.qualified_name)
+                    print('%3d) %s' % (idx, func.qualified_name))
 
                 while True:
                     try:
@@ -782,11 +811,11 @@ class CyBreak(CythonCommand):
                             break_funcs = funcs
                             break
                         elif (result.isdigit() and
-                            0 <= int(result) < len(funcs)):
+                                0 <= int(result) < len(funcs)):
                             break_funcs = [funcs[int(result)]]
                             break
                         else:
-                            print 'Not understood...'
+                            print('Not understood...')
             else:
                 break_funcs = [funcs[0]]
 
@@ -813,18 +842,28 @@ class CyBreak(CythonCommand):
 
     @dont_suppress_errors
     def complete(self, text, word):
-        names = self.cy.functions_by_qualified_name
+        # Filter init-module functions (breakpoints can be set using
+        # modulename:linenumber).
+        names =  [n for n, L in self.cy.functions_by_name.iteritems()
+                        if any(not f.is_initmodule_function for f in L)]
+        qnames = [n for n, f in self.cy.functions_by_qualified_name.iteritems()
+                        if not f.is_initmodule_function]
+
         if parameters.complete_unqualified:
-            names = itertools.chain(names, self.cy.functions_by_name)
+            all_names = itertools.chain(qnames, names)
+        else:
+            all_names = qnames
 
         words = text.strip().split()
-        if words and '.' in words[-1]:
-            lastword = words[-1]
-            compl = [n for n in self.cy.functions_by_qualified_name
-                           if n.startswith(lastword)]
-        else:
+        if not words or '.' not in words[-1]:
+            # complete unqualified
             seen = set(text[:-len(word)].split())
-            return [n for n in names if n.startswith(word) and n not in seen]
+            return [n for n in all_names
+                          if n.startswith(word) and n not in seen]
+
+        # complete qualified name
+        lastword = words[-1]
+        compl = [n for n in qnames if n.startswith(lastword)]
 
         if len(lastword) > len(word):
             # readline sees something (e.g. a '.') as a word boundary, so don't
@@ -835,10 +874,9 @@ class CyBreak(CythonCommand):
         return compl
 
 
-class CythonCodeStepper(CythonCommand, libpython.GenericCodeStepper):
+class CythonInfo(CythonBase, libpython.PythonInfo):
     """
-    Base class for CyStep and CyNext. It implements the interface dictated by
-    libpython.GenericCodeStepper.
+    Implementation of the interface dictated by libpython.LanguageInfo.
     """
 
     def lineno(self, frame):
@@ -848,32 +886,49 @@ class CythonCodeStepper(CythonCommand, libpython.GenericCodeStepper):
         # related code. The C level should be dispatched to the 'step' command.
         if self.is_cython_function(frame):
             return self.get_cython_lineno(frame)
-        else:
-            return libpython.py_step.lineno(frame)
+        return super(CythonInfo, self).lineno(frame)
 
     def get_source_line(self, frame):
         try:
-            line = super(CythonCodeStepper, self).get_source_line(frame)
+            line = super(CythonInfo, self).get_source_line(frame)
         except gdb.GdbError:
             return None
         else:
             return line.strip() or None
 
-    @classmethod
-    def register(cls):
-        return cls(cls.name, stepinto=getattr(cls, 'stepinto', False))
+    def exc_info(self, frame):
+        if self.is_python_function:
+            return super(CythonInfo, self).exc_info(frame)
 
     def runtime_break_functions(self):
         if self.is_cython_function():
             return self.get_cython_function().step_into_functions
+        return ()
 
     def static_break_functions(self):
         result = ['PyEval_EvalFrameEx']
         result.extend(self.cy.functions_by_cname)
         return result
 
+
+class CythonExecutionControlCommand(CythonCommand,
+                                    libpython.ExecutionControlCommandBase):
+
+    @classmethod
+    def register(cls):
+        return cls(cls.name, cython_info)
+
+
+class CyStep(CythonExecutionControlCommand, libpython.PythonStepperMixin):
+    "Step through Cython, Python or C code."
+
+    name = 'cy -step'
+    stepinto = True
+
     def invoke(self, args, from_tty):
-        if not self.is_cython_function() and not self.is_python_function():
+        if self.is_python_function():
+            self.python_step(self.stepinto)
+        elif not self.is_cython_function():
             if self.stepinto:
                 command = 'step'
             else:
@@ -881,24 +936,17 @@ class CythonCodeStepper(CythonCommand, libpython.GenericCodeStepper):
 
             self.finish_executing(gdb.execute(command, to_string=True))
         else:
-            self.step()
+            self.step(stepinto=self.stepinto)
 
 
-class CyStep(CythonCodeStepper):
-    "Step through Cython, Python or C code."
+class CyNext(CyStep):
+    "Step-over Cython, Python or C code."
 
-    name = 'cy step'
-    stepinto = True
-
-
-class CyNext(CythonCodeStepper):
-    "Step-over Python code."
-
-    name = 'cy next'
+    name = 'cy -next'
     stepinto = False
 
 
-class CyRun(CythonCodeStepper):
+class CyRun(CythonExecutionControlCommand):
     """
     Run a Cython program. This is like the 'run' command, except that it
     displays Cython or Python source lines as well
@@ -906,26 +954,26 @@ class CyRun(CythonCodeStepper):
 
     name = 'cy run'
 
-    invoke = CythonCodeStepper.run
+    invoke = CythonExecutionControlCommand.run
 
 
-class CyCont(CyRun):
+class CyCont(CythonExecutionControlCommand):
     """
     Continue a Cython program. This is like the 'run' command, except that it
     displays Cython or Python source lines as well.
     """
 
     name = 'cy cont'
-    invoke = CythonCodeStepper.cont
+    invoke = CythonExecutionControlCommand.cont
 
 
-class CyFinish(CyRun):
+class CyFinish(CythonExecutionControlCommand):
     """
     Execute until the function returns.
     """
     name = 'cy finish'
 
-    invoke = CythonCodeStepper.finish
+    invoke = CythonExecutionControlCommand.finish
 
 
 class CyUp(CythonCommand):
@@ -940,7 +988,7 @@ class CyUp(CythonCommand):
             gdb.execute(self._command, to_string=True)
             while not self.is_relevant_function(gdb.selected_frame()):
                 gdb.execute(self._command, to_string=True)
-        except RuntimeError, e:
+        except RuntimeError as e:
             raise gdb.GdbError(*e.args)
 
         frame = gdb.selected_frame()
@@ -961,7 +1009,7 @@ class CyDown(CyUp):
     _command = 'down'
 
 
-class CySelect(CythonCodeStepper):
+class CySelect(CythonCommand):
     """
     Select a frame. Use frame numbers as listed in `cy backtrace`.
     This command is useful because `cy backtrace` prints a reversed backtrace.
@@ -979,11 +1027,11 @@ class CySelect(CythonCodeStepper):
         while frame.newer():
             frame = frame.newer()
 
-        stackdepth = self._stackdepth(frame)
+        stackdepth = libpython.stackdepth(frame)
 
         try:
             gdb.execute('select %d' % (stackdepth - stackno - 1,))
-        except RuntimeError, e:
+        except RuntimeError as e:
             raise gdb.GdbError(*e.args)
 
 
@@ -998,7 +1046,7 @@ class CyBacktrace(CythonCommand):
     @require_running_program
     def invoke(self, args, from_tty):
         # get the first frame
-        selected_frame = frame = gdb.selected_frame()
+        frame = gdb.selected_frame()
         while frame.older():
             frame = frame.older()
 
@@ -1006,13 +1054,10 @@ class CyBacktrace(CythonCommand):
 
         index = 0
         while frame:
-            is_c = False
-
-            is_relevant = False
             try:
                 is_relevant = self.is_relevant_function(frame)
             except CyGDBError:
-                pass
+                is_relevant = False
 
             if print_all or is_relevant:
                 self.print_stackframe(frame, index)
@@ -1020,8 +1065,6 @@ class CyBacktrace(CythonCommand):
             index += 1
             frame = frame.newer()
 
-        selected_frame.select()
-
 
 class CyList(CythonCommand):
     """
@@ -1033,12 +1076,12 @@ class CyList(CythonCommand):
     command_class = gdb.COMMAND_FILES
     completer_class = gdb.COMPLETE_NONE
 
-    @dispatch_on_frame(c_command='list')
+    # @dispatch_on_frame(c_command='list')
     def invoke(self, _, from_tty):
         sd, lineno = self.get_source_desc()
         source = sd.get_source(lineno - 5, lineno + 5, mark_line=lineno,
                                lex_entire=True)
-        print source
+        print(source)
 
 
 class CyPrint(CythonCommand):
@@ -1072,7 +1115,8 @@ class CyPrint(CythonCommand):
             return []
 
 
-sortkey = lambda (name, value): name.lower()
+sortkey = lambda item: item[0].lower()
+
 
 class CyLocals(CythonCommand):
     """
@@ -1085,7 +1129,13 @@ class CyLocals(CythonCommand):
 
     @dispatch_on_frame(c_command='info locals', python_command='py-locals')
     def invoke(self, args, from_tty):
-        local_cython_vars = self.get_cython_function().locals
+        cython_function = self.get_cython_function()
+
+        if cython_function.is_initmodule_function:
+            self.cy.globals.invoke(args, from_tty)
+            return
+
+        local_cython_vars = cython_function.locals
         max_name_length = len(max(local_cython_vars, key=len))
         for name, cyvar in sorted(local_cython_vars.iteritems(), key=sortkey):
             if self.is_initialized(self.get_cython_function(), cyvar.name):
@@ -1119,13 +1169,13 @@ class CyGlobals(CyLocals):
         max_name_length = max(max_globals_len, max_globals_dict_len)
 
         seen = set()
-        print 'Python globals:'
+        print('Python globals:')
         for k, v in sorted(global_python_dict.iteritems(), key=sortkey):
             v = v.get_truncated_repr(libpython.MAX_OUTPUT_LEN)
             seen.add(k)
-            print '    %-*s = %s' % (max_name_length, k, v)
+            print('    %-*s = %s' % (max_name_length, k, v))
 
-        print 'C globals:'
+        print('C globals:')
         for name, cyvar in sorted(module_globals.iteritems(), key=sortkey):
             if name not in seen:
                 try:
@@ -1138,19 +1188,16 @@ class CyGlobals(CyLocals):
                                              max_name_length, '    ')
 
 
-class CyExec(CythonCommand, libpython.PyExec):
+class EvaluateOrExecuteCodeMixin(object):
     """
-    Execute Python code in the nearest Python or Cython frame.
+    Evaluate or execute Python code in a Cython or Python frame. The 'evalcode'
+    method evaluations Python code, prints a traceback if an exception went
+    uncaught, and returns any return value as a gdb.Value (NULL on exception).
     """
 
-    name = '-cy-exec'
-    command_class = gdb.COMMAND_STACK
-    completer_class = gdb.COMPLETE_NONE
-
     def _fill_locals_dict(self, executor, local_dict_pointer):
         "Fill a remotely allocated dict with values from the Cython C stack"
         cython_func = self.get_cython_function()
-        current_lineno = self.get_cython_lineno()
 
         for name, cyvar in cython_func.locals.iteritems():
             if (cyvar.type == PythonObject and
@@ -1178,28 +1225,21 @@ class CyExec(CythonCommand, libpython.PyExec):
                         raise gdb.GdbError("Unable to execute Python code.")
                 finally:
                     # PyDict_SetItem doesn't steal our reference
-                    executor.decref(pystringp)
+                    executor.xdecref(pystringp)
 
     def _find_first_cython_or_python_frame(self):
         frame = gdb.selected_frame()
         while frame:
             if (self.is_cython_function(frame) or
                 self.is_python_function(frame)):
+                frame.select()
                 return frame
 
             frame = frame.older()
 
         raise gdb.GdbError("There is no Cython or Python frame on the stack.")
 
-    def invoke(self, expr, from_tty):
-        frame = self._find_first_cython_or_python_frame()
-        if self.is_python_function(frame):
-            libpython.py_exec.invoke(expr, from_tty)
-            return
-
-        expr, input_type = self.readcode(expr)
-        executor = libpython.PythonCodeExecutor()
-
+    def _evalcode_cython(self, executor, code, input_type):
         with libpython.FetchAndRestoreError():
             # get the dict of Cython globals and construct a dict in the
             # inferior with Cython locals
@@ -1207,14 +1247,68 @@ class CyExec(CythonCommand, libpython.PyExec):
                 '(PyObject *) PyModule_GetDict(__pyx_m)')
             local_dict = gdb.parse_and_eval('(PyObject *) PyDict_New()')
 
-            cython_function = self.get_cython_function()
-
             try:
                 self._fill_locals_dict(executor,
                                        libpython.pointervalue(local_dict))
-                executor.evalcode(expr, input_type, global_dict, local_dict)
+                result = executor.evalcode(code, input_type, global_dict,
+                                           local_dict)
             finally:
-                executor.decref(libpython.pointervalue(local_dict))
+                executor.xdecref(libpython.pointervalue(local_dict))
+
+        return result
+
+    def evalcode(self, code, input_type):
+        """
+        Evaluate `code` in a Python or Cython stack frame using the given
+        `input_type`.
+        """
+        frame = self._find_first_cython_or_python_frame()
+        executor = libpython.PythonCodeExecutor()
+        if self.is_python_function(frame):
+            return libpython._evalcode_python(executor, code, input_type)
+        return self._evalcode_cython(executor, code, input_type)
+
+
+class CyExec(CythonCommand, libpython.PyExec, EvaluateOrExecuteCodeMixin):
+    """
+    Execute Python code in the nearest Python or Cython frame.
+    """
+
+    name = '-cy-exec'
+    command_class = gdb.COMMAND_STACK
+    completer_class = gdb.COMPLETE_NONE
+
+    def invoke(self, expr, from_tty):
+        expr, input_type = self.readcode(expr)
+        executor = libpython.PythonCodeExecutor()
+        executor.xdecref(self.evalcode(expr, executor.Py_single_input))
+
+
+class CySet(CythonCommand):
+    """
+    Set a Cython variable to a certain value
+
+        cy set my_cython_c_variable = 10
+        cy set my_cython_py_variable = $cy_eval("{'doner': 'kebab'}")
+
+    This is equivalent to
+
+        set $cy_value("my_cython_variable") = 10
+    """
+
+    name = 'cy set'
+    command_class = gdb.COMMAND_DATA
+    completer_class = gdb.COMPLETE_NONE
+
+    @require_cython_frame
+    def invoke(self, expr, from_tty):
+        name_and_expr = expr.split('=', 1)
+        if len(name_and_expr) != 2:
+            raise gdb.GdbError("Invalid expression. Use 'cy set var = expr'.")
+
+        varname, expr = name_and_expr
+        cname = self.cy.cy_cname.invoke(varname.strip())
+        gdb.execute("set %s = %s" % (cname, expr))
 
 
 # Functions
@@ -1263,17 +1357,16 @@ class CyCValue(CyCName):
     @require_cython_frame
     @gdb_function_value_to_unicode
     def invoke(self, cyname, frame=None):
-        try:
+        globals_dict = self.get_cython_globals_dict()
+        cython_function = self.get_cython_function(frame)
+
+        if self.is_initialized(cython_function, cyname):
             cname = super(CyCValue, self).invoke(cyname, frame=frame)
             return gdb.parse_and_eval(cname)
-        except (gdb.GdbError, RuntimeError), e:
-            # variable exists but may not have been initialized yet, or may be
-            # in the globals dict of the Cython module
-            d = self.get_cython_globals_dict()
-            if cyname in d:
-                return d[cyname]._gdbval
-
-            raise gdb.GdbError(str(e))
+        elif cyname in globals_dict:
+            return globals_dict[cyname]._gdbval
+        else:
+            raise gdb.GdbError("Variable %s is not initialized." % cyname)
 
 
 class CyLine(gdb.Function, CythonBase):
@@ -1286,4 +1379,39 @@ class CyLine(gdb.Function, CythonBase):
         return self.get_cython_lineno()
 
 
+class CyEval(gdb.Function, CythonBase, EvaluateOrExecuteCodeMixin):
+    """
+    Evaluate Python code in the nearest Python or Cython frame and return
+    """
+
+    @gdb_function_value_to_unicode
+    def invoke(self, python_expression):
+        input_type = libpython.PythonCodeExecutor.Py_eval_input
+        return self.evalcode(python_expression, input_type)
+
+
+cython_info = CythonInfo()
 cy = CyCy.register()
+cython_info.cy = cy
+
+
+def register_defines():
+    libpython.source_gdb_script(textwrap.dedent("""\
+        define cy step
+        cy -step
+        end
+
+        define cy next
+        cy -next
+        end
+
+        document cy step
+        %s
+        end
+
+        document cy next
+        %s
+        end
+    """) % (CyStep.__doc__, CyNext.__doc__))
+
+register_defines()
