@@ -71,24 +71,31 @@ except NameError:
 WITH_CYTHON = True
 CY3_DIR = None
 
-from distutils.dist import Distribution
-from distutils.core import Extension
 from distutils.command.build_ext import build_ext as _build_ext
 from distutils import sysconfig
-distutils_distro = Distribution()
 
 
-if sys.platform == 'win32':
-    # TODO: Figure out why this hackery (see http://thread.gmane.org/gmane.comp.python.cython.devel/8280/).
-    config_files = distutils_distro.find_config_files()
-    try: config_files.remove('setup.cfg')
-    except ValueError: pass
-    distutils_distro.parse_config_files(config_files)
+def get_distutils_distro(_cache=[]):
+    if _cache:
+        return _cache[0]
+    # late import to accomodate for setuptools override
+    from distutils.dist import Distribution
+    distutils_distro = Distribution()
+
+    if sys.platform == 'win32':
+        # TODO: Figure out why this hackery (see http://thread.gmane.org/gmane.comp.python.cython.devel/8280/).
+        config_files = distutils_distro.find_config_files()
+        try: config_files.remove('setup.cfg')
+        except ValueError: pass
+        distutils_distro.parse_config_files(config_files)
+
+        cfgfiles = distutils_distro.find_config_files()
+        try: cfgfiles.remove('setup.cfg')
+        except ValueError: pass
+        distutils_distro.parse_config_files(cfgfiles)
+    _cache.append(distutils_distro)
+    return distutils_distro
 
-    cfgfiles = distutils_distro.find_config_files()
-    try: cfgfiles.remove('setup.cfg')
-    except ValueError: pass
-    distutils_distro.parse_config_files(cfgfiles)
 
 EXT_DEP_MODULES = {
     'tag:numpy' : 'numpy',
@@ -841,7 +848,7 @@ class CythonCompileTestCase(unittest.TestCase):
         cwd = os.getcwd()
         os.chdir(workdir)
         try:
-            build_extension = build_ext(distutils_distro)
+            build_extension = build_ext(get_distutils_distro())
             build_extension.include_dirs = INCLUDE_DIRS[:]
             if incdir:
                 build_extension.include_dirs.append(incdir)
@@ -861,10 +868,12 @@ class CythonCompileTestCase(unittest.TestCase):
 
             related_files = self.related_files(test_directory, module)
             self.copy_files(test_directory, workdir, related_files)
+
+            from distutils.core import Extension
             extension = Extension(
                 module,
-                sources = self.source_files(workdir, module, related_files),
-                extra_compile_args = ext_compile_flags,
+                sources=self.source_files(workdir, module, related_files),
+                extra_compile_args=ext_compile_flags,
                 **extra_extension_args
                 )
 
@@ -1951,7 +1960,9 @@ def runtests(options, cmd_args, coverage=None):
 
     try:
         import jedi
-    except ImportError:
+        if list(map(int, re.findall('[0-9]+', jedi.__version__ or '0'))) < [0, 8, 1]:
+            raise ImportError
+    except (ImportError, AttributeError, TypeError):
         exclude_selectors.append(RegExSelector('Jedi'))
 
     if options.exclude:
